/*****************************************************************************
** FILE IDENTIFICATION
**
**   Name:          projections.h
**   Purpose:       Header file for Projections class
**   Programmer:    Kevin Rosenberg
**   Date Started:  July 1, 1984
**
**  This is part of the CTSim program
**  Copyright (C) 1983-2000 Kevin Rosenberg
**
**  $Id: projections.h,v 1.18 2000/12/16 06:12:47 kevin Exp $
**
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License (version 2) as
**  published by the Free Software Foundation.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
******************************************************************************/

#ifndef PROJECTIONS_H
#define PROJECTIONS_H

class Scanner;
class DetectorArray;
class Array2dFileLabel;


// Projections
class Projections
{
 public:
  Projections (const int nView, const int nDet);
  Projections (const Scanner& scanner);
  Projections ();
  ~Projections ();

  void initFromScanner (const Scanner& scanner);

  void printProjectionData (int startView, int endView);
  void printProjectionData ();
  void printScanInfo (std::ostringstream& os) const;

  bool read (const std::string& fname);
  bool read (const char* fname);
  bool write (const char* fname);
  bool write (const std::string& fname);
  bool detarrayRead (fnetorderstream& fs, DetectorArray& darray, const int view_num);
  bool detarrayWrite (fnetorderstream& fs, const DetectorArray& darray, const int view_num);

  bool reconstruct (ImageFile& im, const char* const filterName, double filt_param, const char* const filterMethodName, const int zeropad, const char* frequencyFilterName, const char* const interpName, int interp_param, const char* const backprojName, const int trace) const;

  void setNView (int nView);  // used in MPI to restrict # of views
  void setRotInc (double rotInc) { m_rotInc = rotInc;}
  void setDetInc (double detInc) { m_detInc = detInc;}
  void setPhmLen (double phmLen) { m_phmLen = phmLen;}
  void setCalcTime (double calcTime) {m_calcTime = calcTime;}
  void setRemark (const char* remark) {m_remark = remark; m_label.setLabelString(remark);}
  void setRemark (const std::string& remark) {setRemark(remark.c_str());}

  double detStart() const {return m_detStart;}
  double rotStart() const {return m_rotStart;}
  double calcTime() const {return m_calcTime;}
  double phmLen() const {return m_phmLen;}
  const char*  remark() const {return m_remark.c_str();}
  double detInc() const {return m_detInc;}
  double rotInc() const {return m_rotInc;}
  int nDet() const {return m_nDet;}
  int nView() const {return m_nView;}
  int geometry() const {return m_geometry;}
  double focalLength() const {return m_focalLength;}
  double fieldOfView() const {return m_fieldOfView;}

  const std::string& getFilename() const {return m_filename;}
  Array2dFileLabel& getLabel() {return m_label;}
  const Array2dFileLabel& getLabel() const {return m_label;}

  DetectorArray& getDetectorArray (const int iview)
      { return (*m_projData[iview]); }
  
  const DetectorArray& getDetectorArray (const int iview) const
      { return (*m_projData[iview]); }

  static bool copyHeader (const char* const filename, std::ostream& os);
  static bool copyHeader (const std::string& filename, std::ostream& os);

  static bool copyViewData (const char* const filename, std::ostream& os, int startView, int endView);
  static bool copyViewData (const std::string& filename, std::ostream& os, int startView, int endView);
  
 private:
  int m_headerSize;		// Size of disk file header 
  int m_geometry;		// Geometry of scanner 
  class DetectorArray **m_projData;	// Pointer to array of detarray_st pointers 
  std::string m_remark;	        // description of raysum data 
  int m_nDet;			// number of detectors in array 
  int m_nView;			// number of rotated views 
  double m_calcTime;		// time required to calculate raysums 
  double m_rotStart;		// starting view rotation
  double m_rotInc;		// angle between rotations 
  double m_detStart;		// distance of beginning detector to center phantom
  double m_detInc;		// increment between detectors 
  double m_phmLen;		// Length of phantom edge (phm is square) 
  double m_focalLength;
  double m_fieldOfView;
  kuint32 m_year;                   // Creation date & time
  kuint32 m_month;
  kuint32 m_day;
  kuint32 m_hour;
  kuint32 m_minute;
  kuint32 m_second;
  std::string m_filename;
  Array2dFileLabel m_label;

  const static kuint16 m_signature;

  bool headerWrite (fnetorderstream& fs);
  bool headerRead (fnetorderstream& fs);
  void newProjData ();
  void deleteProjData ();

  void init (const int nView, const int nDet);

  // prevent default methods
  Projections& operator= (const Projections& rhs);   // assignment
  Projections(const Projections& rhs);               // copy
};


#endif
