/*****************************************************************************
** FILE IDENTIFICATION
**
**	Name:	      imagefile.cpp
**      Purpose:      Imagefile classes
**	Programmer:   Kevin Rosenberg
**	Date Started: June 2000
**
**  This is part of the CTSim program
**  Copyright (C) 1983-2000 Kevin Rosenberg
**
**  $Id: imagefile.cpp,v 1.27 2000/12/29 20:15:37 kevin Exp $
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License (version 2) as
**  published by the Free Software Foundation.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
******************************************************************************/

#include "ct.h"


F32Image::F32Image (int nx, int ny, int dataType)
: Array2dFile (nx, ny, sizeof(kfloat32), Array2dFile::PIXEL_FLOAT32, dataType)
{
}

F32Image::F32Image (void)
: Array2dFile()
{
	setPixelFormat (Array2dFile::PIXEL_FLOAT32);
	setPixelSize (sizeof(kfloat32));
  setDataType (Array2dFile::DATA_TYPE_REAL);
}

F64Image::F64Image (int nx, int ny, int dataType)
: Array2dFile (nx, ny, sizeof(kfloat64), Array2dFile::PIXEL_FLOAT64, dataType)
{
}

F64Image::F64Image (void)
: Array2dFile ()
{
	setPixelFormat (PIXEL_FLOAT64);
	setPixelSize (sizeof(kfloat64));
  setDataType (Array2dFile::DATA_TYPE_REAL);
}

void 
ImageFile::filterResponse (const char* const domainName, double bw, const char* const filterName, double filt_param)
{
	int hx = (m_nx - 1) / 2;
	int hy = (m_ny - 1) / 2;
	ImageFileArray v = getArray();
	SignalFilter filter (filterName, domainName, bw, filt_param);
	
	for (int i = -hx; i <= hx; i++) {
		for (int j = -hy; j <= hy; j++) {
      double r = ::sqrt (i * i + j * j);
			
			v[i+hx][j+hy] = filter.response (r);
		}
	}
}

int
ImageFile::display (void) const
{
    double pmin, pmax;
	
    getMinMax (pmin, pmax);
	
    return (displayScaling (1, pmin, pmax));
}

int 
ImageFile::displayScaling (const int scale, const ImageFileValue pmin, const ImageFileValue pmax) const
{
    int nx = m_nx;
    int ny = m_ny;
    ImageFileArrayConst v = getArray();
    if (v == NULL || nx == 0 || ny == 0)
		return 0;
	
#if HAVE_G2_H
    int* pPens = new int [nx * ny * scale * scale ];
	
    double view_scale = 255 / (pmax - pmin);
    int id_X11 = g2_open_X11 (nx * scale, ny * scale);
    int grayscale[256];
    for (int i = 0; i < 256; i++) {
		double cval = i / 255.;
		grayscale[i] = g2_ink (id_X11, cval, cval, cval);
    }
	
    for (int iy = ny - 1; iy >= 0; iy--) {
		int iRowPos = ((ny - 1 - iy) * scale) * (nx * scale);
		for (int ix = 0; ix < nx; ix++) {
			int cval = static_cast<int>((v[ix][iy] - pmin) * view_scale);
			if (cval < 0)  
				cval = 0;
			else if (cval > 255) 
				cval = 255;
			for (int sy = 0; sy < scale; sy++)
				for (int sx = 0; sx < scale; sx++)
					pPens[iRowPos+(sy * nx * scale)+(sx + (ix * scale))] = grayscale[cval];
		}
    }
	
    g2_image (id_X11, 0., 0., nx * scale, ny * scale, pPens);
	
    delete pPens;
    return (id_X11);
#else
    return 0;
#endif
}



// ImageFile::comparativeStatistics    Calculate comparative stats
//
// OUTPUT
//   d   Normalized root mean squared distance measure
//   r   Normalized mean absolute distance measure
//   e   Worst case distance measure
//
// REFERENCES
//  G.T. Herman, Image Reconstruction From Projections, 1980

bool
ImageFile::comparativeStatistics (const ImageFile& imComp, double& d, double& r, double& e) const
{
    if (imComp.nx() != m_nx && imComp.ny() != m_ny) {
		sys_error (ERR_WARNING, "Image sizes differ [ImageFile::comparativeStatistics]");
		return false;
    }
    ImageFileArrayConst v = getArray();
    if (v == NULL || m_nx == 0 || m_ny == 0)
		return false;
	
    ImageFileArrayConst vComp = imComp.getArray();
	
    double myMean = 0.;
    for (unsigned int ix = 0; ix < m_nx; ix++) {
		for (unsigned int iy = 0; iy < m_ny; iy++) {
			myMean += v[ix][iy];
		}
    }
    myMean /= (m_nx * m_ny);
	
    double sqErrorSum = 0.;
    double absErrorSum = 0.;
    double sqDiffFromMeanSum = 0.;
    double absValueSum = 0.;
    for (unsigned int ix2 = 0; ix2 < m_nx; ix2++) {
		for (unsigned int iy = 0; iy < m_ny; iy++) {
			double diff = v[ix2][iy] - vComp[ix2][iy];
			sqErrorSum += diff * diff;
			absErrorSum += fabs(diff);
			double diffFromMean = v[ix2][iy] - myMean;
			sqDiffFromMeanSum += diffFromMean * diffFromMean;
			absValueSum += fabs(v[ix2][iy]);
		}
    }
	
    d = ::sqrt (sqErrorSum / sqDiffFromMeanSum);
    r = absErrorSum / absValueSum;
	
    int hx = m_nx / 2;
    int hy = m_ny / 2;
    double eMax = -1;
    for (int ix3 = 0; ix3 < hx; ix3++) {
		for (int iy = 0; iy < hy; iy++) {
			double avgPixel = 0.25 * (v[2*ix3][2*iy] + v[2*ix3+1][2*iy] + v[2*ix3][2*iy+1] + v[2*ix3+1][2*iy+1]);
			double avgPixelComp = 0.25 * (vComp[2*ix3][2*iy] + vComp[2*ix3+1][2*iy] + vComp[2*ix3][2*iy+1] + vComp[2*ix3+1][2*iy+1]);
			double error = fabs (avgPixel - avgPixelComp);
			if (error > eMax)
				eMax = error;
		}
    }
	
    e = eMax;
	
    return true;
}


bool
ImageFile::printComparativeStatistics (const ImageFile& imComp, std::ostream& os) const
{
	double d, r, e;
	
	if (comparativeStatistics (imComp, d, r, e)) {
		os << "  Normalized root mean squared distance (d): " << d << std::endl;
		os << "      Normalized mean absolute distance (r): " << r << std::endl;
		os << "Worst case distance (2x2 pixel average) (e): " << e << std::endl;
		return true;
	}
	return false;
}


void
ImageFile::printStatistics (std::ostream& os) const
{
    double min, max, mean, mode, median, stddev;
	
    statistics (min, max, mean, mode, median, stddev);
    if (dataType() == Array2dFile::DATA_TYPE_COMPLEX)
      os << "Real Component Statistics" << std::endl;

    os << "   min: " << min << std::endl;
    os << "   max: " << max << std::endl;
    os << "  mean: " << mean << std::endl;
    os << "  mode: " << mode << std::endl;
    os << "median: " << median << std::endl;
    os << "stddev: " << stddev << std::endl;

    if (dataType() == Array2dFile::DATA_TYPE_COMPLEX) {
      statistics (getImaginaryArray(), min, max, mean, mode, median, stddev);
      os << std::endl << "Imaginary Component Statistics" << std::endl;
    os << "   min: " << min << std::endl;
    os << "   max: " << max << std::endl;
    os << "  mean: " << mean << std::endl;
    os << "  mode: " << mode << std::endl;
    os << "median: " << median << std::endl;
    os << "stddev: " << stddev << std::endl;
    }
}


void
ImageFile::statistics (double& min, double& max, double& mean, double& mode, double& median, double& stddev) const
{
  ImageFileArrayConst v = getArray();
  statistics (v, min, max, mean, mode, median, stddev);
}


void
ImageFile::statistics (ImageFileArrayConst& v, double& min, double& max, double& mean, double& mode, double& median, double& stddev) const
{
    int nx = m_nx;
    int ny = m_ny;
    
    if (v == NULL || nx == 0 || ny == 0)
		return;

	std::vector<double> vecImage;
	int iVec = 0;
	vecImage.resize (nx * ny);
    for (int ix = 0; ix < nx; ix++) {
		for (int iy = 0; iy < ny; iy++)
			vecImage[iVec++] = v[ix][iy];
	}

	vectorNumericStatistics (vecImage, nx * ny, min, max, mean, mode, median, stddev);
}

void
ImageFile::getMinMax (double& min, double& max) const
{
    int nx = m_nx;
    int ny = m_ny;
    ImageFileArrayConst v = getArray();
    
    if (v == NULL || nx == 0 || ny == 0)
		return;
	
    min = v[0][0];
    max = v[0][0];
    for (int ix = 0; ix < nx; ix++) {
		for (int iy = 0; iy < ny; iy++) {
			if (v[ix][iy] > max)
				max = v[ix][iy];
			if (v[ix][iy] < min)
				min = v[ix][iy];
		}
    }
}

bool
ImageFile::convertRealToComplex ()
{
  if (dataType() != Array2dFile::DATA_TYPE_REAL)
    return false;

  if (! reallocRealToComplex())
    return false;

  ImageFileArray vImag = getImaginaryArray();
  for (unsigned int ix = 0; ix < m_nx; ix++) {
    ImageFileColumn vCol = vImag[ix];
    for (unsigned int iy = 0; iy < m_ny; iy++)
      *vCol++ = 0;
  }

  return true;
}

bool
ImageFile::convertComplexToReal ()
{
  if (dataType() != Array2dFile::DATA_TYPE_COMPLEX)
    return false;

  ImageFileArray vReal = getArray();
  ImageFileArray vImag = getImaginaryArray();
  for (unsigned int ix = 0; ix < m_nx; ix++) {
    ImageFileColumn vRealCol = vReal[ix];
    ImageFileColumn vImagCol = vImag[ix];
    for (unsigned int iy = 0; iy < m_ny; iy++) {
      std::complex<double> c (*vRealCol, *vImagCol);
      *vRealCol++ = std::abs (c);
      vImagCol++;
    }
  }

  return reallocComplexToReal();
}

bool
ImageFile::subtractImages (const ImageFile& rRHS, ImageFile& result) const
{
  if (m_nx != rRHS.nx() || m_ny != rRHS.ny() || m_nx != result.nx() || m_ny != result.ny()) {
    sys_error (ERR_WARNING, "Difference sizes of images [ImageFile::subtractImage]");
    return false;
  }

  ImageFileArrayConst vLHS = getArray();
  ImageFileArrayConst vRHS = rRHS.getArray();
  ImageFileArray vResult = result.getArray();

  for (unsigned int ix = 0; ix < m_nx; ix++) {
    ImageFileColumnConst in1 = vLHS[ix];
    ImageFileColumnConst in2 = vRHS[ix];
    ImageFileColumn out = vResult[ix];
    for (unsigned int iy = 0; iy < m_ny; iy++)
        *out++ = *in1++ - *in2++;
  }

    return true;
}

bool
ImageFile::addImages (const ImageFile& rRHS, ImageFile& result) const
{
  if (m_nx != rRHS.nx() || m_ny != rRHS.ny() || m_nx != result.nx() || m_ny != result.ny()) {
    sys_error (ERR_WARNING, "Difference sizes of images [ImageFile::subtractImage]");
    return false;
  }

  ImageFileArrayConst vLHS = getArray();
  ImageFileArrayConst vRHS = rRHS.getArray();
  ImageFileArray vResult = result.getArray();

  for (unsigned int ix = 0; ix < m_nx; ix++) {
    ImageFileColumnConst in1 = vLHS[ix];
    ImageFileColumnConst in2 = vRHS[ix];
    ImageFileColumn out = vResult[ix];
    for (unsigned int iy = 0; iy < m_ny; iy++)
        *out++ = *in1++ + *in2++;
  }

    return true;
}

bool
ImageFile::multiplyImages (const ImageFile& rRHS, ImageFile& result) const
{
  if (m_nx != rRHS.nx() || m_ny != rRHS.ny() || m_nx != result.nx() || m_ny != result.ny()) {
    sys_error (ERR_WARNING, "Difference sizes of images [ImageFile::subtractImage]");
    return false;
  }

  ImageFileArrayConst vLHS = getArray();
  ImageFileArrayConst vRHS = rRHS.getArray();
  ImageFileArray vResult = result.getArray();

  for (unsigned int ix = 0; ix < m_nx; ix++) {
    ImageFileColumnConst in1 = vLHS[ix];
    ImageFileColumnConst in2 = vRHS[ix];
    ImageFileColumn out = vResult[ix];
    for (unsigned int iy = 0; iy < m_ny; iy++)
        *out++ = *in1++ * *in2++;
  }

    return true;
}

bool
ImageFile::divideImages (const ImageFile& rRHS, ImageFile& result) const
{
  if (m_nx != rRHS.nx() || m_ny != rRHS.ny() || m_nx != result.nx() || m_ny != result.ny()) {
    sys_error (ERR_WARNING, "Difference sizes of images [ImageFile::subtractImage]");
    return false;
  }

  ImageFileArrayConst vLHS = getArray();
  ImageFileArrayConst vRHS = rRHS.getArray();
  ImageFileArray vResult = result.getArray();

  for (unsigned int ix = 0; ix < m_nx; ix++) {
    ImageFileColumnConst in1 = vLHS[ix];
    ImageFileColumnConst in2 = vRHS[ix];
    ImageFileColumn out = vResult[ix];
    for (unsigned int iy = 0; iy < m_ny; iy++) {
      if (*in2 != 0.)
        *out++ = *in1++ / *in2++;
      else
        *out++ = 0;
    }
  }

    return true;
}


bool
ImageFile::invertPixelValues (ImageFile& result) const
{
  if (m_nx != result.nx() || m_ny != result.ny()) {
    sys_error (ERR_WARNING, "Difference sizes of images [ImageFile::invertPixelValues]");
    return false;
  }

  ImageFileArrayConst vLHS = getArray();
  ImageFileArray vResult = result.getArray();

  for (unsigned int ix = 0; ix < m_nx; ix++) {
    ImageFileColumnConst in = vLHS[ix];
    ImageFileColumn out = vResult[ix];
    for (unsigned int iy = 0; iy < m_ny; iy++)
        *out++ = - *in++;
  }

  return true;
}

bool
ImageFile::sqrt (ImageFile& result) const
{
  if (m_nx != result.nx() || m_ny != result.ny()) {
    sys_error (ERR_WARNING, "Difference sizes of images [ImageFile::invertPixelValues]");
    return false;
  }

  ImageFileArrayConst vLHS = getArray();
  ImageFileArray vResult = result.getArray();

  for (unsigned int ix = 0; ix < m_nx; ix++) {
    ImageFileColumnConst in = vLHS[ix];
    ImageFileColumn out = vResult[ix];
    for (unsigned int iy = 0; iy < m_ny; iy++)
      if (*in < 0)
        *out++ = -::sqrt(-*in++);
      else
        *out++ = ::sqrt(*in++);
  }

  return true;
}

bool
ImageFile::log (ImageFile& result) const
{
  if (m_nx != result.nx() || m_ny != result.ny()) {
    sys_error (ERR_WARNING, "Difference sizes of images [ImageFile::invertPixelValues]");
    return false;
  }

  ImageFileArrayConst vLHS = getArray();
  ImageFileArray vResult = result.getArray();

  for (unsigned int ix = 0; ix < m_nx; ix++) {
    ImageFileColumnConst in = vLHS[ix];
    ImageFileColumn out = vResult[ix];
    for (unsigned int iy = 0; iy < m_ny; iy++)
      if (*in <= 0)
        *out++ = 0;
      else
        *out++ = ::log(*in++);
  }

  return true;
}

bool
ImageFile::exp (ImageFile& result) const
{
  if (m_nx != result.nx() || m_ny != result.ny()) {
    sys_error (ERR_WARNING, "Difference sizes of images [ImageFile::invertPixelValues]");
    return false;
  }

  ImageFileArrayConst vLHS = getArray();
  ImageFileArray vResult = result.getArray();

  for (unsigned int ix = 0; ix < m_nx; ix++) {
    ImageFileColumnConst in = vLHS[ix];
    ImageFileColumn out = vResult[ix];
    for (unsigned int iy = 0; iy < m_ny; iy++)
      *out++ = ::exp (*in++);
  }

  return true;
}

bool
ImageFile::inverseFourier (ImageFile& result) const
{
  if (m_nx != result.nx() || m_ny != result.ny()) {
    sys_error (ERR_WARNING, "Difference sizes of images [ImageFile::invertPixelValues]");
    return false;
  }

  return true;
}

bool
ImageFile::fourier (ImageFile& result) const
{
  if (m_nx != result.nx() || m_ny != result.ny()) {
    sys_error (ERR_WARNING, "Difference sizes of images [ImageFile::invertPixelValues]");
    return false;
  }

  if (result.dataType() == Array2dFile::DATA_TYPE_REAL) {
    if (! result.reallocRealToComplex ())
      return false;
  }

  ImageFileArrayConst vLHS = getArray();
  ImageFileArray vRealResult = result.getArray();
  ImageFileArray vImagResult = result.getImaginaryArray();

  unsigned int ix, iy;
  double* pY = new double [m_ny];
  std::complex<double>** complexOut = new std::complex<double>* [m_nx];
  for (ix = 0; ix < m_nx; ix++)
    complexOut[ix] = new std::complex<double> [m_ny];

  for (ix = 0; ix < m_nx; ix++) {
    for (iy = 0; iy < m_ny; iy++)
      pY[iy] = vLHS[ix][iy];
    ProcessSignal::finiteFourierTransform (pY, complexOut[ix], m_ny,  ProcessSignal::FORWARD);
  }
  delete pY;

  std::complex<double>* pX = new std::complex<double> [m_nx];
  std::complex<double>* complexOutCol = new std::complex<double> [m_nx];
  for (iy = 0; iy < m_ny; iy++) {
    for (ix = 0; ix < m_nx; ix++)
      pX[ix] = complexOut[ix][iy];
    ProcessSignal::finiteFourierTransform (pX, complexOutCol, m_nx, ProcessSignal::FORWARD);
    for (ix = 0; ix < m_nx; ix++)
      complexOut[ix][iy] = complexOutCol[ix];
  }
  delete [] pX;

  for (ix = 0; ix < m_nx; ix++)
    ProcessSignal::shuffleFourierToNaturalOrder (complexOut[ix], m_ny);

  
  for (iy = 0; iy < m_ny; iy++) {
    for (ix = 0; ix < m_nx; ix++)
      complexOutCol[ix] = complexOut[ix][iy];
    ProcessSignal::shuffleFourierToNaturalOrder (complexOutCol, m_nx);;
    for (ix = 0; ix < m_nx; ix++)
      complexOut[ix][iy] = complexOutCol[ix];

  }
  delete [] complexOutCol;

  for (ix = 0; ix < m_nx; ix++)
    for (iy = 0; iy < m_ny; iy++) {
      vRealResult[ix][iy] = complexOut[ix][iy].real();
      vImagResult[ix][iy] = complexOut[ix][iy].imag();
    }

  for (ix = 0; ix < m_nx; ix++)
    delete [] complexOut[ix];

  delete [] complexOut;

  return true;
}

bool
ImageFile::magnitude (ImageFile& result) const
{
  if (m_nx != result.nx() || m_ny != result.ny()) {
    sys_error (ERR_WARNING, "Difference sizes of images [ImageFile::invertPixelValues]");
    return false;
  }

  ImageFileArray vReal = getArray();
  ImageFileArray vImag = NULL;
  if (dataType() == Array2dFile::DATA_TYPE_COMPLEX)
    vImag = getImaginaryArray();

  ImageFileArray vRealResult = result.getArray();
  if (result.dataType() == Array2dFile::DATA_TYPE_COMPLEX) {
    ImageFileArray vImagResult = result.getImaginaryArray();
    for (unsigned int ix = 0; ix < m_nx; ix++)
      for (unsigned int iy = 0; iy < m_ny; iy++)
        vImagResult[ix][iy] = 0;
  }

  for (unsigned int ix = 0; ix < m_nx; ix++)
    for (unsigned int iy = 0; iy < m_ny; iy++) {
      if (dataType() == Array2dFile::DATA_TYPE_COMPLEX) 
        vRealResult[ix][iy] = ::sqrt (vReal[ix][iy] * vReal[ix][iy] + vImag[ix][iy] * vImag[ix][iy]);
      else
        vRealResult[ix][iy] = vReal[ix][iy];
    }

  return true;
}

bool
ImageFile::phase (ImageFile& result) const
{
  if (m_nx != result.nx() || m_ny != result.ny()) {
    sys_error (ERR_WARNING, "Difference sizes of images [ImageFile::invertPixelValues]");
    return false;
  }

  ImageFileArray vReal = getArray();
  ImageFileArray vImag = NULL;
  if (dataType() == Array2dFile::DATA_TYPE_COMPLEX)
    vImag = getImaginaryArray();

  ImageFileArray vRealResult = result.getArray();
  if (result.dataType() == Array2dFile::DATA_TYPE_COMPLEX) {
    ImageFileArray vImagResult = result.getImaginaryArray();
    for (unsigned int ix = 0; ix < m_nx; ix++)
      for (unsigned int iy = 0; iy < m_ny; iy++)
        vImagResult[ix][iy] = 0;
  }

  for (unsigned int ix = 0; ix < m_nx; ix++)
    for (unsigned int iy = 0; iy < m_ny; iy++) {
      if (dataType() == Array2dFile::DATA_TYPE_COMPLEX) 
        vRealResult[ix][iy] = ::atan (vImag[ix][iy] / vReal[ix][iy]);
      else
        vRealResult[ix][iy] = vReal[ix][iy];
    }

  return true;
}

bool
ImageFile::square (ImageFile& result) const
{
  if (m_nx != result.nx() || m_ny != result.ny()) {
    sys_error (ERR_WARNING, "Difference sizes of images [ImageFile::invertPixelValues]");
    return false;
  }

  ImageFileArrayConst vLHS = getArray();
  ImageFileArray vResult = result.getArray();

  for (unsigned int ix = 0; ix < m_nx; ix++) {
    ImageFileColumnConst in = vLHS[ix];
    ImageFileColumn out = vResult[ix];
    for (unsigned int iy = 0; iy < m_ny; iy++) {
        *out++ = *in * *in;
        in++;
    }
  }

  return true;
}


void 
ImageFile::writeImagePGM (const char *outfile, int nxcell, int nycell, double densmin, double densmax)
{
	FILE *fp;
	int nx = m_nx;
	int ny = m_ny;
	ImageFileArray v = getArray();
	
	unsigned char* rowp = new unsigned char [nx * nxcell];
	
	if ((fp = fopen (outfile, "wb")) == NULL)
		return;
	
	fprintf(fp, "P5\n");
	fprintf(fp, "%d %d\n", nx, ny);
	fprintf(fp, "255\n");
	
	for (int irow = ny - 1; irow >= 0; irow--) {
		for (int icol = 0; icol < nx; icol++) {
			int pos = icol * nxcell;
			double dens = (v[icol][irow] - densmin) / (densmax - densmin);
			dens = clamp (dens, 0., 1.);
			for (int p = pos; p < pos + nxcell; p++) {
				rowp[p] = static_cast<unsigned int> (dens * 255.);
			}
		}
		for (int ir = 0; ir < nycell; ir++) {
			for (int ic = 0; ic < nx * nxcell; ic++) 
				fputc( rowp[ic], fp );
		}
	}
	
	delete rowp;
	fclose(fp);
}

void 
ImageFile::writeImagePGMASCII (const char *outfile, int nxcell, int nycell, double densmin, double densmax)
{
	FILE *fp;
	int nx = m_nx;
	int ny = m_ny;
	ImageFileArray v = getArray();
	
	unsigned char* rowp = new unsigned char [nx * nxcell];
	
	if ((fp = fopen (outfile, "wb")) == NULL)
		return;
	
	fprintf(fp, "P2\n");
	fprintf(fp, "%d %d\n", nx, ny);
	fprintf(fp, "255\n");
	
	for (int irow = ny - 1; irow >= 0; irow--) {
		for (int icol = 0; icol < nx; icol++) {
			int pos = icol * nxcell;
			double dens = (v[icol][irow] - densmin) / (densmax - densmin);
			dens = clamp (dens, 0., 1.);
			for (int p = pos; p < pos + nxcell; p++) {
				rowp[p] = static_cast<unsigned int> (dens * 255.);
			}
		}
		for (int ir = 0; ir < nycell; ir++) {
			for (int ic = 0; ic < nx * nxcell; ic++) 
				fprintf(fp, "%d ", rowp[ic]);
			fprintf(fp, "\n");
		}
	}
	
	delete rowp;
	fclose(fp);
}


#ifdef HAVE_PNG
void 
ImageFile::writeImagePNG (const char *outfile, int bitdepth, int nxcell, int nycell, double densmin, double densmax)
{
	FILE *fp;
	png_structp png_ptr;
	png_infop info_ptr;
	double max_out_level = (1 << bitdepth) - 1;
	int nx = m_nx;
	int ny = m_ny;
	ImageFileArray v = getArray();
	
	unsigned char* rowp = new unsigned char [nx * nxcell * (bitdepth / 8)];
	
	if ((fp = fopen (outfile, "wb")) == NULL)
		return;
	
	png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (! png_ptr)
		return;
	
	info_ptr = png_create_info_struct(png_ptr);
	if (! info_ptr) {
		png_destroy_write_struct(&png_ptr, (png_infopp) NULL);
		fclose(fp);
		return;
	}
	
	if (setjmp(png_ptr->jmpbuf)) {
		png_destroy_write_struct(&png_ptr, &info_ptr);
		fclose(fp);
		return;
	}
	
	png_init_io(png_ptr, fp);
	
	png_set_IHDR(png_ptr, info_ptr, nx * nxcell, ny * nycell, bitdepth, PNG_COLOR_TYPE_GRAY, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_DEFAULT);
	
	png_write_info(png_ptr, info_ptr);
	for (int irow = ny - 1; irow >= 0; irow--) {
		png_bytep row_pointer = rowp;
		
		for (int icol = 0; icol < nx; icol++) {
			int pos = icol * nxcell;
			double dens = (v[icol][irow] - densmin) / (densmax - densmin);
			dens = clamp (dens, 0., 1.);
			unsigned int outval = static_cast<unsigned int> (dens * max_out_level);
			
			for (int p = pos; p < pos + nxcell; p++) {
				if (bitdepth == 8)
					rowp[p] = outval;
				else {
					int rowpos = p * 2;
					rowp[rowpos] = (outval >> 8) & 0xFF;
					rowp[rowpos+1] = (outval & 0xFF);
				}
			}
		}
		for (int ir = 0; ir < nycell; ir++)
			png_write_rows (png_ptr, &row_pointer, 1);
	}
	
	png_write_end(png_ptr, info_ptr);
	png_destroy_write_struct(&png_ptr, &info_ptr);
	delete rowp;
	
	fclose(fp);
}
#endif

#ifdef HAVE_GD
#include "gd.h"
static const int N_GRAYSCALE=256;

void
ImageFile::writeImageGIF (const char *outfile, int nxcell, int nycell, double densmin, double densmax)
{
	gdImagePtr gif;
	FILE *out;
	int gs_indices[N_GRAYSCALE];
	int nx = m_nx;
	int ny = m_ny;
	ImageFileArray v = getArray();
	
	unsigned char rowp [nx * nxcell];
	if (rowp == NULL)
		return;
	
	gif = gdImageCreate(nx * nxcell, ny * nycell);
	for (int i = 0; i < N_GRAYSCALE; i++)
		gs_indices[i] = gdImageColorAllocate(gif, i, i, i);
	
	int lastrow = ny * nycell - 1;
	for (int irow = 0; irow < ny; irow++) {
		int rpos = irow * nycell;
		for (int ir = rpos; ir < rpos + nycell; ir++) {
			for (int icol = 0; icol < nx; icol++) {
				int cpos = icol * nxcell;
				double dens = (v[icol][irow] - densmin) / (densmax - densmin);
				dens = clamp(dens, 0., 1.);
				for (int ic = cpos; ic < cpos + nxcell; ic++) {
					rowp[ic] = (unsigned int) (dens * (double) (N_GRAYSCALE - 1));
					gdImageSetPixel(gif, ic, lastrow - ir, gs_indices[rowp[ic]]);
				}
			}
		}
	}
	
	if ((out = fopen(outfile,"w")) == NULL) {
		sys_error(ERR_FATAL, "Error opening output file %s for writing", outfile);
		return (1);
	}
	gdImageGif(gif,out);
	fclose(out);
	gdImageDestroy(gif);
}
#endif

