/*****************************************************************************
** FILE IDENTIFICATION
**
**   Name:	   projections.cpp         Projection data classes
**   Programmer:   Kevin Rosenberg
**   Date Started: Aug 84
**
**  This is part of the CTSim program
**  Copyright (C) 1983-2000 Kevin Rosenberg
**
**  $Id: projections.cpp,v 1.33 2000/12/17 22:30:09 kevin Exp $
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License (version 2) as
**  published by the Free Software Foundation.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
******************************************************************************/

#include "ct.h"

const kuint16 Projections::m_signature = ('P'*256 + 'J');

/* NAME
 *    Projections		Constructor for projections matrix storage 
 *
 * SYNOPSIS
 *    proj = projections_create (filename, nView, nDet)
 *    Projections& proj		Allocated projections structure & matrix
 *    int nView			Number of rotated view
 *    int nDet			Number of detectors
 *
 */

Projections::Projections (const Scanner& scanner)
  : m_projData(0)
{
  initFromScanner (scanner);
}


Projections::Projections (const int nView, const int nDet)
  : m_projData(0)
{
  init (nView, nDet);
}

Projections::Projections (void)
  : m_projData(0)
{
  init (0, 0);
}

Projections::~Projections (void)
{
  deleteProjData();
}


void
Projections::init (const int nView, const int nDet)
{
  m_label.setLabelType (Array2dFileLabel::L_HISTORY);
  m_nView = nView;
  m_nDet = nDet;
  newProjData ();
}

void
Projections::initFromScanner (const Scanner& scanner)
{
  m_label.setLabelType (Array2dFileLabel::L_HISTORY);
  deleteProjData();
  init (scanner.nView(), scanner.nDet());

  m_phmLen = scanner.phmLen();
  m_rotInc = scanner.rotInc();
  m_detInc = scanner.detInc();
  m_geometry = scanner.geometry();
  m_focalLength = scanner.focalLength();
  m_fieldOfView = scanner.fieldOfView();
  m_rotStart = 0;
  m_detStart =  -(scanner.detLen() / 2);
#if 0
  if (m_geometry == Scanner::GEOMETRY_EQUILINEAR) {
    m_detInc /= 2;
	std::cout << "Kludge: detInc /= 2 in Projections::initFromScanner" << endl;
  }
#endif
}

void
Projections::setNView (int nView)  // used by MPI to reduce # of views
{
  deleteProjData();
  init (nView, m_nDet);
}

// NAME
// newProjData

void 
Projections::newProjData (void)
{
  if (m_projData)
    sys_error(ERR_WARNING, "m_projData != NULL [newProjData]");

  if (m_nView > 0 && m_nDet) {
    m_projData = new DetectorArray* [m_nView];

    for (int i = 0; i < m_nView; i++)
      m_projData[i] = new DetectorArray (m_nDet);
  }
}


/* NAME
 *   projections_free			Free memory allocated to projections
 *
 * SYNOPSIS
 *   projections_free(proj)
 *   Projections& proj				Projectionss to be deallocated
 */

void 
Projections::deleteProjData (void)
{
  if (m_projData != NULL) {
    for (int i = 0; i < m_nView; i++)
      delete m_projData[i];

    delete m_projData;
    m_projData = NULL;
  }
}


/* NAME
 *    Projections::headerWwrite         Write data header for projections file
 *
 */

bool 
Projections::headerWrite (fnetorderstream& fs)
{
  kuint16 _hsize = m_headerSize;
  kuint16 _signature = m_signature;
  kuint32 _nView = m_nView;
  kuint32 _nDet = m_nDet;
  kuint32 _geom = m_geometry;
  kuint16 _remarksize = m_remark.length();
  kuint16 _year = m_year;
  kuint16 _month = m_month;
  kuint16 _day = m_day;
  kuint16 _hour = m_hour;
  kuint16 _minute = m_minute;
  kuint16 _second = m_second;

  kfloat64 _calcTime = m_calcTime;
  kfloat64 _rotStart = m_rotStart;
  kfloat64 _rotInc = m_rotInc;
  kfloat64 _detStart = m_detStart;
  kfloat64 _detInc = m_detInc;
  kfloat64 _phmLen = m_phmLen;
  kfloat64 _fieldOfView = m_fieldOfView;
  kfloat64 _focalLength = m_focalLength;

  fs.seekp(0);
  if (! fs)
    return false;

  fs.writeInt16 (_hsize);
  fs.writeInt16 (_signature);
  fs.writeInt32 (_nView);
  fs.writeInt32 (_nDet);
  fs.writeInt32 (_geom);
  fs.writeFloat64 (_calcTime);
  fs.writeFloat64 (_rotStart);
  fs.writeFloat64 (_rotInc);
  fs.writeFloat64 (_detStart);
  fs.writeFloat64 (_detInc);
  fs.writeFloat64 (_phmLen);
  fs.writeFloat64 (_focalLength);
  fs.writeFloat64 (_fieldOfView);
  fs.writeInt16 (_year);
  fs.writeInt16 (_month);
  fs.writeInt16 (_day);
  fs.writeInt16 (_hour);
  fs.writeInt16 (_minute);
  fs.writeInt16 (_second);
  fs.writeInt16 (_remarksize);
  fs.write (m_remark.c_str(), _remarksize);

  m_headerSize = fs.tellp();
  _hsize = m_headerSize;
  fs.seekp(0);
  fs.writeInt16 (_hsize);
  if (! fs)
      return false;
  
  return true;
}

/* NAME
 *    projections_read_header         Read data header for projections file
 *
 */
bool
Projections::headerRead (fnetorderstream& fs)
{
  kuint16 _hsize, _signature, _year, _month, _day, _hour, _minute, _second, _remarksize = 0;
  kuint32 _nView, _nDet, _geom;
  kfloat64 _calcTime, _rotStart, _rotInc, _detStart, _detInc, _phmLen, _focalLength, _fieldOfView;
  
  fs.seekg(0);
  if (! fs)
      return false;

  fs.readInt16 (_hsize);
  fs.readInt16 (_signature);
  fs.readInt32 (_nView);
  fs.readInt32 (_nDet);
  fs.readInt32 (_geom);
  fs.readFloat64 (_calcTime);
  fs.readFloat64 (_rotStart);
  fs.readFloat64 (_rotInc);
  fs.readFloat64 (_detStart);
  fs.readFloat64 (_detInc);
  fs.readFloat64 (_phmLen);
  fs.readFloat64 (_focalLength);
  fs.readFloat64 (_fieldOfView);
  fs.readInt16 (_year);
  fs.readInt16 (_month);
  fs.readInt16 (_day);
  fs.readInt16 (_hour);
  fs.readInt16 (_minute);
  fs.readInt16 (_second);
  fs.readInt16 (_remarksize);

  if (! fs) {
      sys_error (ERR_SEVERE, "Error reading header information , _remarksize=%d [projections_read_header]", _remarksize);
      return false;
  }

  if (_signature != m_signature) {
    sys_error (ERR_SEVERE, "File %s does not have a valid projection file signature", m_filename.c_str());
    return false;
  }

  char* pszRemarkStorage = new char [_remarksize+1];
  fs.read (pszRemarkStorage, _remarksize);
  if (! fs) {
    sys_error (ERR_SEVERE, "Error reading remark, _remarksize = %d", _remarksize);
    return false;
  }
  pszRemarkStorage[_remarksize] = 0;
  m_remark = pszRemarkStorage;
  delete pszRemarkStorage;

  off_t _hsizeread = fs.tellg();
  if (!fs || _hsizeread != _hsize) {
      sys_error (ERR_WARNING, "File header size read %ld != file header size stored %ld [read_projections_header]\n_remarksize=%ld", (long int) _hsizeread, _hsize, _remarksize);
      return false;
  }
  
  m_headerSize = _hsize;
  m_nView = _nView;
  m_nDet = _nDet;
  m_geometry = _geom;
  m_calcTime = _calcTime;
  m_rotStart = _rotStart;
  m_rotInc = _rotInc;
  m_detStart = _detStart;
  m_detInc = _detInc;
  m_phmLen = _phmLen;
  m_focalLength = _focalLength;
  m_fieldOfView = _fieldOfView;
  m_year = _year;
  m_month = _month;
  m_day = _day;
  m_hour = _hour;
  m_minute = _minute;
  m_second = _second;

  m_label.setLabelType (Array2dFileLabel::L_HISTORY);
  m_label.setLabelString (m_remark);
  m_label.setCalcTime (m_calcTime);
  m_label.setDateTime (m_year, m_month, m_day, m_hour, m_minute, m_second);

  return true;
}

bool
Projections::read (const std::string& filename)
{
  return read (filename.c_str());
}


bool
Projections::read (const char* filename)
{
  m_filename = filename;
#ifdef MSVC
  frnetorderstream fileRead (m_filename.c_str(), std::ios::in | std::ios::binary);
#else
  frnetorderstream fileRead (m_filename.c_str(), std::ios::in | std::ios::binary | std::ios::nocreate);
#endif

  if (fileRead.fail())
    return false;

  if (! headerRead (fileRead))
    return false;

  deleteProjData ();
  newProjData();

  for (int i = 0; i < m_nView; i++) {
    if (! detarrayRead (fileRead, *m_projData[i], i))
      break;
  }

  fileRead.close();
  return true;
}


bool 
Projections::copyViewData (const std::string& filename, std::ostream& os, int startView, int endView)
{
    return copyViewData (filename.c_str(), os, startView, endView);
}

bool 
Projections::copyViewData (const char* const filename, std::ostream& os, int startView, int endView)
{
  frnetorderstream is (filename, ios::in | ios::binary);
  kuint16 sizeHeader, signature;
  kuint32 _nView, _nDet;

  is.readInt16 (sizeHeader);
  is.readInt16 (signature);
  is.readInt32 (_nView);
  is.readInt32 (_nDet);
  int nView = _nView;
  int nDet = _nDet;

  if (signature != m_signature) {
    sys_error (ERR_FATAL, "Illegal signature in projection file %s", filename);
    return false;
  }

  if (startView < 0)
      startView = 0;
  if (startView > nView - 1)
      startView = nView;
  if (endView < 0 || endView > nView - 1)
      endView = nView - 1;

  if (startView > endView) { // swap if start > end
      int tempView = endView;
      endView = startView;
      startView = tempView;
  }

  int sizeView = 8 /* view_angle */ + 4 /* nDet */ + (4 * nDet);
  unsigned char* pViewData = new unsigned char [sizeView];

  for (int i = startView; i <= endView; i++) {
      is.seekg (sizeHeader + i * sizeView);
      is.read (reinterpret_cast<char*>(pViewData), sizeView);
      os.write (reinterpret_cast<char*>(pViewData), sizeView);
      if (is.fail() || os.fail())
	  break;
  }

  delete pViewData;
  if (is.fail()) 
    sys_error (ERR_FATAL, "Error reading projection file");
  if (os.fail()) 
    sys_error (ERR_FATAL, "Error writing projection file");
    
  return (! (is.fail() | os.fail()));
}

bool 
Projections::copyHeader (const std::string& filename, std::ostream& os)
{
    return copyHeader (filename.c_str(), os);
}

bool
Projections::copyHeader (const char* const filename, std::ostream& os)
{
  frnetorderstream is (filename, std::ios::in | std::ios::binary);
  kuint16 sizeHeader, signature;
  is.readInt16 (sizeHeader);
  is.readInt16 (signature);
  is.seekg (0);
  if (signature != m_signature) {
    sys_error (ERR_FATAL, "Illegal signature in projection file %s", filename);
    return false;
  }

  unsigned char* pHdrData = new unsigned char [sizeHeader];
  is.read (reinterpret_cast<char*>(pHdrData), sizeHeader);
  if (is.fail()) {
      sys_error (ERR_FATAL, "Error reading header");
      return false;
  }

  os.write (reinterpret_cast<char*>(pHdrData), sizeHeader);
  if (os.fail()) {
      sys_error (ERR_FATAL, "Error writing header");
      return false;
  }

  return true;
}

bool
Projections::write (const std::string& filename)
{
  return write (filename.c_str());
}

bool
Projections::write (const char* filename)
{
  frnetorderstream fs (filename, std::ios::out | std::ios::binary | std::ios::trunc | std::ios::ate);
  m_filename = filename;
  if (! fs) {
    sys_error (ERR_SEVERE, "Error opening file %s for output [projections_create]", filename);
    return false;
  }

#ifdef HAVE_TIME
  time_t t = time(NULL);
  tm* lt = localtime(&t);
  m_year = lt->tm_year;
  m_month = lt->tm_mon;
  m_day = lt->tm_mday;
  m_hour = lt->tm_hour;
  m_minute = lt->tm_min;
  m_second = lt->tm_sec;
#endif

  if (! headerWrite (fs))
      return false;

  if (m_projData != NULL) {
    for (int i = 0; i < m_nView; i++) {
      if (! detarrayWrite (fs, *m_projData[i], i))
	break;
    }
  }
  if (! fs)
    return false;

 fs.close();

  return true;
}

/* NAME
 *   detarrayRead		Read a Detector Array structure from the disk
 *
 * SYNOPSIS
 *   detarrayRead (proj, darray, view_num)
 *   DETARRAY *darray		Detector array storage location to be filled
 *   int      view_num		View number to read
 */

bool
Projections::detarrayRead (fnetorderstream& fs, DetectorArray& darray, const int iview)
{
  const int detval_bytes = darray.nDet() * sizeof(kfloat32);
  const int detheader_bytes = sizeof(kfloat64) /* view_angle */ + sizeof(kint32) /* nDet */;
  const int view_bytes = detheader_bytes + detval_bytes;
  const off_t start_data = m_headerSize + (iview * view_bytes);
  DetectorValue* detval_ptr = darray.detValues();  
  kfloat64 view_angle;
  kuint32 nDet;

  fs.seekg (start_data);

  fs.readFloat64 (view_angle);
  fs.readInt32 (nDet);
  darray.setViewAngle (view_angle);
  //  darray.setNDet ( nDet);
  
  for (unsigned int i = 0; i < nDet; i++) {
      kfloat32 detval;
      fs.readFloat32 (detval);
      detval_ptr[i] = detval;
  }
  if (! fs)
    return false;

  return true;
}


/* NAME
 *   detarrayWrite			Write detector array data to the disk
 *
 * SYNOPSIS
 *   detarrayWrite (darray, view_num)
 *   DETARRAY *darray			Detector array data to be written
 *   int      view_num			View number to write
 *
 * DESCRIPTION
 *       This routine writes the detarray data from the disk sequentially to
 *    the file that was opened with open_projections().  Data is written in
 *    binary format.
 */

bool
Projections::detarrayWrite (fnetorderstream& fs, const DetectorArray& darray, const int iview)
{
  const int detval_bytes = darray.nDet() * sizeof(float);
  const int detheader_bytes = sizeof(kfloat64) /* view_angle */ + sizeof(kint32) /* nDet */;
  const int view_bytes = detheader_bytes + detval_bytes;
  const off_t start_data = m_headerSize + (iview * view_bytes);
  const DetectorValue* const detval_ptr = darray.detValues();  
  kfloat64 view_angle = darray.viewAngle();
  kuint32 nDet = darray.nDet();
  
  fs.seekp (start_data);
  if (! fs) {
    sys_error (ERR_SEVERE, "Error seeking detectory array [detarrayWrite]");
    return false;
  }

  fs.writeFloat64 (view_angle);
  fs.writeInt32 (nDet);

  for (unsigned int i = 0; i < nDet; i++) {
    kfloat32 detval = detval_ptr[i];
    fs.writeFloat32 (detval);
  }

  if (! fs)
    return (false);

  return true;
}

/* NAME
 *   printProjectionData			Print projections data
 *
 * SYNOPSIS
 *   printProjectionData ()
 */

void
Projections::printProjectionData ()
{
  printProjectionData (0, nView() - 1);
}

void
Projections::printProjectionData (int startView, int endView)
{
  printf("Projections Data\n\n");
  printf("Description: %s\n", m_remark.c_str());
  printf("Geometry: %s\n", Scanner::convertGeometryIDToName (m_geometry));
  printf("nView       = %8d           nDet = %8d\n", m_nView, m_nDet);
  printf("focalLength = %8.4f  fieldOfView = %8.4f\n", m_focalLength, m_fieldOfView);
  printf("rotStart    = %8.4f       rotInc = %8.4f\n", m_rotStart, m_rotInc);
  printf("detStart    = %8.4f       detInc = %8.4f\n", m_detStart, m_detInc);
  if (m_projData != NULL) {
    if (startView < 0)
      startView = 0;
    if (startView > m_nView - 1)
      startView = m_nView - 1;
    if (endView > m_nView - 1)
      endView = m_nView - 1;
    for (int ir = startView; ir <= endView - 1; ir++) {
      printf("View %d: angle %f\n", ir, m_projData[ir]->viewAngle());
      DetectorValue* detval = m_projData[ir]->detValues();
      for (int id = 0; id < m_projData[ir]->nDet(); id++)
	printf("%8.4f  ", detval[id]);
      printf("\n");
    }
  }
}

void 
Projections::printScanInfo (std::ostringstream& os) const
{
  os << "Number of detectors: " << m_nDet << "\n";
  os << "    Number of views: " << m_nView<< "\n";
  os << "             Remark: " << m_remark.c_str()<< "\n";
  os << "           Geometry: " << Scanner::convertGeometryIDToName (m_geometry)<< "\n";
  os << "       Focal Length: " << m_focalLength<< "\n";
  os << "      Field Of View: " << m_fieldOfView<< "\n";
  os << "             phmLen: " << m_phmLen<< "\n";
  os << "           detStart: " << m_detStart<< "\n";
  os << "             detInc: " << m_detInc<< "\n";
  os << "           rotStart: " << m_rotStart<< "\n";
  os << "             rotInc: " << m_rotInc<< "\n";
}



